import datetime
import math

from pint import UndefinedUnitError

from kamzik3 import units
from kamzik3.constants import *


def _base_unit(unit):
    try:
        prefix, base_unit, suffix = next(units.parse_unit_name(unit))
    except (UndefinedUnitError, AttributeError, StopIteration):
        prefix, base_unit, suffix = None, unit, None
    return prefix, base_unit, suffix


def get_attribute_unit_range(attribute):
    prefix, base_unit, suffix = _base_unit(attribute[UNIT])

    if base_unit in ("ampere", "volt", "meter", "degree", "radian", "second", "hertz", "percent"):
        sorted_prefixes = (
            'y', 'z', 'a', 'f', 'p', 'n', 'u', 'm', "", 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'
        )
    elif base_unit in ("degC", "degF", "K"):
        return ["degC", "degF", "K"]
    elif base_unit in ("bar", "Pa", "atm"):
        return ["mbar", "bar", "Pa", "kPa", "atm"]
    else:
        sorted_prefixes = (
            'y', 'z', 'a', 'f', 'p', 'n', 'u', 'm', 'c', 'd', "", 'da', 'h', 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'
        )
    try:
        converter = units._prefixes[prefix]._converter
        if converter == 1:
            scale = 1
        else:
            scale = converter.scale
    except KeyError:
        scale = 1

    try:
        min_unit = scale * 10 ** -attribute[DECIMALS]
        max_unit = scale * abs(attribute[MAX])
    except TypeError:
        raise TypeError

    unit_range = []
    for prefix in sorted_prefixes:
        converter = units._prefixes[prefix]._converter
        if converter == 1:
            scale = 1
        else:
            scale = converter.scale
        if min_unit <= scale <= max_unit:
            unit_range.append(units.get_symbol("{}{}".format(prefix, base_unit)))

    return unit_range


def get_decimals_diff(old_unit, new_unit):
    old_scale = math.log(units.get_base_units(old_unit)[0], 10)
    new_scale = math.log(units.get_base_units(new_unit)[0], 10)
    return int(round(old_scale - new_scale))


def device_units(device, attribute, value):
    try:
        value = value.replace("%", "percent")
    except AttributeError:
        pass

    device_unit = device.get_attribute(attribute)[UNIT]
    # try:
    converted_value = units.Quantity(value)
    # No unit for value
    if converted_value.unitless:
        # Return in value in device units
        try:
            return units.Quantity(float(units.Quantity(value, device_unit).m), device_unit)
        except AttributeError:
            return units.Quantity(float(units.Quantity(value).m))
    else:
        # Return converted value
        return converted_value.to(device_unit)
    # except DimensionalityError:
    #     # Values are not convertable among each other
    #     return units.Quantity(value, device_unit)


MAX_SECONDS = 2 ** 31 - 1


def get_scaled_time_duration(seconds):
    if seconds < 0: return
    if seconds > MAX_SECONDS:
        seconds = MAX_SECONDS

    d = datetime.datetime(1, 1, 1) + datetime.timedelta(seconds=seconds)

    if seconds < 60:
        milliseconds = int(d.microsecond / 1000)
        if milliseconds > 0:
            time_estimate = "{} Sec {} MSec".format(d.second, milliseconds)
        else:
            time_estimate = "{} Sec".format(d.second)
    elif 60 <= seconds < 3600:
        time_estimate = "{} Min {} Sec".format(d.minute, d.second)
    elif 3600 <= seconds < 24 * 3600:
        time_estimate = "{} Hour {} Min".format(d.hour, d.minute)
    elif 24 * 3600 <= seconds < 24 * 3600 * 28:
        time_estimate = "{} Day {} Hour {} Min".format(d.day - 1, d.hour, d.minute)
    elif 24 * 3600 * 28 <= seconds < 24 * 3600 * 28 * 12:
        time_estimate = "{} Month {} Day {} Hour".format(d.month, d.day - 1, d.hour)
    else:
        time_estimate = "{} Year {} Month {} Day".format(d.year, d.month, d.day - 1)

    return time_estimate


def seconds_to_datetime(seconds):
    if seconds > MAX_SECONDS:
        seconds = MAX_SECONDS

    return datetime.datetime.fromtimestamp(seconds).strftime("%d.%m.%Y %H:%M:%S")



from ctypes import CDLL

# from msl.loadlib import LoadLibrary, Server32
from kamzik3.devices.attribute import Attribute
from kamzik3.constants import *
from kamzik3 import DeviceError
from kamzik3.devices.device import Device

# class MyServer(Server32):
#     """A wrapper around a 32-bit C++ library, 'cpp_lib32.dll', that has an 'add' function."""
#
#     def __init__(self, host, port, quiet, **kwargs):
#         # Load the 'cpp_lib32' shared-library file using ctypes.CDLL.
#         super(MyServer, self).__init__('cpp_lib32.dll', 'cdll', host, port, quiet)
#
#     def add(self, a, b):
#         # The Server32 class has a 'lib' property that is a reference to the ctypes.CDLL object.
#         # The shared library’s 'add' function takes two integers as inputs and returns the sum.
#         return self.lib.add(a, b)

class DeviceOsLibrary(Device):

    def __init__(self, os_library, device_id=None, config=None):
        self.os_library = os_library
        self.lib = None
        super().__init__(device_id, config)
        self.connect()

    def connect(self, *args):
        """
        Call only this function to connect devices to port / socket / library / ...
        :param args: connect attributes
        """
        self.connecting = True
        self.handle_configuration_event()
        try:
            self.lib = CDLL(self.os_library)
            # self.lib = LoadLibrary(self.os_library)
        except OSError as e:
            raise DeviceError("Error loading library {}. {}".format(self.os_library, e))
        self.connected = True
        self.connecting = False
        self.set_status(STATUS_IDLE)

    def _init_attributes(self):
        Device._init_attributes(self)
        self.attributes.update({
            ATTR_LIBRARY: Attribute(self.os_library, readonly=True),
        })
    


import collections
import math
from threading import Lock
from time import time

import pint
from numpy import dtype, float16, float32, float64, uint16, uint32, uint64, int16, int32, int64, uint8, int8, ndarray, \
    bool
from pint import UndefinedUnitError

from kamzik3 import units
from kamzik3.constants import *


# TODO: Deal with numpy arrays
# TODO: Implement option to force attribute to change it's value even if current one is the same
#       It's needed for some casses like counter
# TODO: Remove current workarounds introduced for Karabo bridge

class SetFunction(object):

    def __init__(self, attribute, callback=None):
        self.attribute = attribute
        self.callback = callback

    def __enter__(self):
        self.attribute.with_set_function = True
        self.attribute.set_callback = self.callback
        return self.attribute

    def __exit__(self, value_type, value, traceback):
        self.attribute.with_set_function = False
        self.attribute.set_callback = None


class Attribute(dict):
    attribute_keys = {VALUE, TYPE, UNIT, READONLY, DESCRIPTION}
    integer_types = (int8, int16, int32, int64, uint8, uint16, uint32, uint64)
    float_types = (float16, float32, float64)
    unsigned_types = (uint8, uint16, uint32, uint64)
    with_set_function = False
    last_update_time = time()
    numerical = False
    set_token = 0
    set_callback = None

    def attribute_copy(self):
        return Attribute(**self)

    def __init__(self, default_value=None, min_value=None, max_value=None,
                 unit=None, set_function=None, default_type=u"str", decimals=0,
                 readonly=False, description=None, set_value_when_set_function=True,
                 tolerance=(0, 0), factor=1, offset=0, **kwargs):
        dict.__init__(self)
        self.set_function = set_function
        self.update_lock = Lock()
        self.default_type = default_type
        self.offset = offset
        self.factor = factor
        # Set attribute value flag when using set function.
        # This flag prevents change of attribute's value before value is actually set on device side.
        self.set_value_when_set_function = set_value_when_set_function
        self._callbacks = {}
        if unit is None:
            unit = u""
        self[UNIT] = unit
        if kwargs.get(TYPE, False) is not False:
            default_type = kwargs.get(TYPE)
        if default_type == TYPE_LIST:
            self[TYPE] = TYPE_LIST
        elif isinstance(default_type, (list, collections.abc.KeysView)):
            self[TYPE] = TYPE_LIST
            self[TYPE_LIST_VALUES] = list(default_type)
        elif isinstance(default_type, ndarray):
            self[TYPE] = TYPE_ARRAY
        else:
            default_type = dtype(default_type)
            self[TYPE] = dtype(default_type).name
            if default_type in self.integer_types:
                self[DECIMALS] = 0
                if min_value is None and default_type in self.unsigned_types:
                    self[MIN] = 0
                else:
                    self[MIN] = int(min_value) if min_value is not None else -math.inf
                self[MAX] = int(max_value) if max_value is not None else math.inf
                self.numerical = True
            elif default_type in self.float_types:
                self[DECIMALS] = decimals
                self[MIN] = float(min_value) if min_value is not None else -math.inf
                self[MAX] = float(max_value) if max_value is not None else math.inf
                self.numerical = True
        self[TOLERANCE] = [float(tolerance[0]), float(tolerance[1])]
        self[READONLY] = readonly
        self[DESCRIPTION] = description
        self[VALUE] = default_value
        self[SETPOINT] = None
        self.update(kwargs)

    def __setitem__(self, k, v):
        if self.update_lock.locked():
            return

        if k == FACTOR:
            self.factor = v
        elif k == OFFSET:
            self.offset = v
        elif k == VALUE:
            if isinstance(v, units.Quantity):
                try:
                    v = v.to(self[UNIT]).m
                except AttributeError:
                    v = v.m
            if self.set_function and self.with_set_function:
                self[SETPOINT] = v
                setpoint = self.remove_offset_factor(v)
                # Release set function flag
                self.with_set_function = False
                # Call set function
                if hasattr(self.set_function, "callback"):
                    self.set_token = self.set_function(setpoint, callback=self.set_callback)
                else:
                    # We don't have callback in a function, so just call it plain without callback
                    self.set_token = self.set_function(setpoint)
                if not self.set_value_when_set_function:
                    return
            else:
                v = self.apply_offset_factor(v)
        elif k in (MIN, MAX):
            v = self.apply_offset_factor(v)
        elif k == TOLERANCE:
            # Make sure that tolerance is float
            v = [float(t) for t in v]
        elif k == UNIT and v is not None:
            if v == "%":
                v = u"percent"
            if v not in ("", None):
                try:
                    # Try to get unit name from pint
                    units.get_name(v)
                except pint.errors.UndefinedUnitError:
                    # Unit does not exists so it's added to the pint definitions
                    units.define(u"{}=1".format(v))
                    units.derived_units.append(v)

        # Store actual value into dictionary
        super(Attribute, self).__setitem__(k, v)

        # Call all relevant callbacks
        for callback, (max_update_rate, updated_at) in self._callbacks.copy().items():
            if k != VALUE or max_update_rate is None or time() - updated_at > max_update_rate:
                callback(k, v)
                try:
                    self._callbacks[callback][1] = time()
                except KeyError:
                    # Callback no longer exists, skipping
                    pass

        self.last_update_time = time()

    def __getitem__(self, k):
        if k == FACTOR:
            return self.factor
        elif k == OFFSET:
            return self.offset
        return super(Attribute, self).__getitem__(k)

    @staticmethod
    def from_dict(dictionary):
        """
        Create Attribute from provided dictionary.
        :param dictionary: dict
        :return: Attribute
        """
        if isinstance(dictionary, dict):
            if Attribute.is_attribute(dictionary):
                if dictionary[UNIT] not in ("", None):
                    try:
                        units.get_name(dictionary[UNIT])
                    except pint.errors.UndefinedUnitError:
                        units.define(u"{}=1".format(dictionary[UNIT]))
                        units.derived_units.append(dictionary[UNIT])
                return Attribute(**dictionary)
            else:
                for key, attribute in dictionary.items():
                    dictionary[key] = Attribute.from_dict(attribute)
            return dictionary
        else:
            return dictionary

    @staticmethod
    def is_attribute(dictionary):
        """
        Check if provided dictionary is convertible to Attribute.
        :param dictionary: dict
        :return: bool
        """
        try:
            return Attribute.attribute_keys.issubset(dictionary.keys())
        except (AttributeError, TypeError, ValueError):
            return False

    def attach_callback(self, callback, max_update_rate=None):
        """
        Attach callback to Attribute.
        :param callback: callable function
        :param max_update_rate: max update rate in ms
        :return: None
        """
        assert callable(callback)

        if callback not in self._callbacks:
            if max_update_rate is not None:
                max_update_rate *= 1e-3
            self._callbacks[callback] = [max_update_rate, time()]

    def detach_callback(self, callback):
        """
        Detach callback from attribute.
        :param callback: callable function
        :return: None
        """
        assert callable(callback)
        try:
            del self._callbacks[callback]
        except KeyError:
            pass  # Callback no longer exists

    def apply_offset_factor(self, value=None):
        """
        Apply offset and factor to attribute value.
        :param value: numerical value
        :return: mixed
        """
        if value is not None and self.numerical:
            return (value + self.offset) * self.factor
        elif self.default_type is bool:
            return self.factor == value
        else:
            return value

    def remove_offset_factor(self, value=None):
        """
        Remove offset and factor from attribute value.
        :param value: numerical value
        :return: mixed
        """
        if value is not None and self.numerical:
            return (value / self.factor) - self.offset
        elif self.default_type is bool:
            return self.factor == value
        else:
            return value

    def read_and_reset_token(self):
        token = self.set_token
        self.set_token = 0
        return token

    @staticmethod
    def list_attribute(attribute):
        """
        Create list from input attribute.
        Many methods relies on the fact that attribute is stored as list.
        :param attribute: mixed
        :return: list
        """
        if isinstance(attribute, list):
            return attribute
        elif isinstance(attribute, tuple):
            return list(attribute)
        else:
            return [attribute]

    def convert_units(self, value):
        """
        Convert input value to current attribute unit.
        :param value: units.Quantity
        :return: value or units.Quantity
        """
        assert isinstance(value, units.Quantity)
        if value.dimensionless:
            if self[UNIT] is None:
                return value
            else:
                return units.Quantity("{} {}".format(value.m, self[UNIT]))

        return value.to(self[UNIT])

    def within_limits(self, value):
        """
        Check if value is within limits.
        :param value:  units.Quantity
        :return: bool
        """
        assert isinstance(value, units.Quantity)
        return self.minimum() <= value <= self.maximum()

    def value(self):
        if self.numerical:
            try:
                return units.Quantity(self[VALUE], self[UNIT])
            except (UndefinedUnitError, AttributeError):
                return units.Quantity(self[VALUE])
            except TypeError:
                return self[VALUE]
        else:
            return self[VALUE]

    def minimum(self):
        if self.numerical:
            try:
                return units.Quantity(self[MIN], self[UNIT])
            except (UndefinedUnitError, AttributeError):
                return units.Quantity(self[MIN])
            except TypeError:
                return self[VALUE]
        else:
            return self[MIN]

    def maximum(self):
        if self.numerical:
            try:
                return units.Quantity(self[MAX], self[UNIT])
            except (UndefinedUnitError, AttributeError):
                return units.Quantity(self[MAX])
            except TypeError:
                return self[VALUE]
        else:
            return self[MAX]

    def negative_tolerance(self):
        if self.numerical:
            try:
                return units.Quantity(self[TOLERANCE][0], self[UNIT])
            except (UndefinedUnitError, AttributeError):
                return units.Quantity(self[TOLERANCE][0])
            except TypeError:
                return self[VALUE]
        else:
            return self[TOLERANCE][0]

    def positive_tolerance(self):
        if self.numerical:
            try:
                return units.Quantity(self[TOLERANCE][1], self[UNIT])
            except (UndefinedUnitError, AttributeError):
                return units.Quantity(self[TOLERANCE][1])
            except TypeError:
                return self[VALUE]
        else:
            return self[TOLERANCE][1]

    def offset(self):
        if self.numerical:
            try:
                return units.Quantity(self[OFFSET], self[UNIT])
            except (UndefinedUnitError, AttributeError):
                return units.Quantity(self[OFFSET])
            except TypeError:
                return self[VALUE]
        else:
            return self[OFFSET]

    def setpoint(self):
        if self.numerical:
            try:
                return units.Quantity(self[SETPOINT], self[UNIT])
            except (UndefinedUnitError, AttributeError):
                return units.Quantity(self[SETPOINT])
            except TypeError:
                return self[SETPOINT]
        else:
            return self[SETPOINT]

    def unit(self):
        return self[UNIT]



import os
import time

from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.devices.device import Device


class DeviceFileSync(Device):

    def __init__(self, filepath, device_id=None, config=None):
        self.filepath = filepath
        super(DeviceFileSync, self).__init__(device_id, config)
        self._expose_methods_to_clients()
        self.connect()

    def _init_attributes(self):
        super(DeviceFileSync, self)._init_attributes()

        self.attributes.update({
            ATTR_FILEPATH: Attribute(set_function=self.set_filepath, readonly=True),
            ATTR_CONTENT: Attribute(set_function=self.set_content),
        })

    def handle_configuration(self):

        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.set_attribute((ATTR_FILEPATH, VALUE), self.filepath)
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        _finish_configuration()

    def set_filepath(self, value):
        if not os.path.exists(value):
            raise DeviceError("File {} does not exists".format(value))
        else:
            with open(value, "r") as fp:
                file_content = "".join(fp.readlines())
                self.set_value(ATTR_CONTENT, file_content)

    def set_content(self, value):
        file_path = self[ATTR_FILEPATH][VALUE]
        with open(file_path, "w") as fp:
            fp.write(value)
            self.set_value(ATTR_CONTENT, value)



import asynchat
import asyncore
import copy
import logging
import socket
import time
from collections import OrderedDict
from threading import Thread, Event, Lock

from kamzik3 import DeviceClientError
from kamzik3.constants import *


class PortReadLoop(Thread):

    def __init__(self):
        super(PortReadLoop, self).__init__()
        self.map = []
        self.stopped = Event()

    def add_device(self, device):
        if device not in self.map:
            self.map.append(device)
            return True
        else:
            return False

    def remove_device(self, device):
        if device in self.map:
            self.map.remove(device)
            return True
        else:
            return False

    def run(self):
        while not self.stopped.wait(0.01):
            for device in self.map:
                try:
                    device.handle_read()
                except IOError:
                    device.handle_response_error(u"Port read error")

    def stop(self):
        self.stopped.set()


class DevicePoller(Thread):
    min_poll = 10  # Minimum polling interval default is 10 ms

    def __init__(self):
        super(DevicePoller, self).__init__()
        self.stopped = Event()
        self.lock = Lock()
        self.connect_lock = Lock()
        self.connecting_devices = []
        self.commands_buffer = {}
        self.logger = logging.getLogger(u"Console.DevicePoller")
        self.polling_schedule = OrderedDict()
        self.schedule = []

    def run(self):
        self.logger.info(u"Starting devices poller Thread")
        schedule_at = self.min_poll
        check_connection_at = 1000
        min_poll_tick = self.min_poll * 1e-3

        while not self.stopped.wait(timeout=min_poll_tick):
            if check_connection_at <= 0:
                for device in self.commands_buffer.keys():
                    device.is_alive()

            with self.lock:
                for pollAt in self.schedule:
                    if pollAt > schedule_at:
                        break
                    elif schedule_at % pollAt == 0:
                        for device, attributes in self.polling_schedule[pollAt].items():
                            if device.accepting_commands():
                                self.commands_buffer[device] += attributes
                for device, commands in copy.copy(self.commands_buffer).items():
                    if commands and device.accepting_commands():
                        self.commands_buffer[device] = device.send_command(commands)
                        device.set_value(ATTR_BUFFERED_COMMANDS, len(self.commands_buffer[device]))

            # Check connection only @ 1Hz
            if check_connection_at <= 0:
                self.check_devices_connection_timeout()
                check_connection_at = 1000
            else:
                check_connection_at -= self.min_poll

            if not self.schedule or schedule_at >= self.schedule[-1]:
                schedule_at = self.min_poll
            else:
                schedule_at += self.min_poll

    def stop(self):
        self.logger.info(u"Stopping devices poller Thread")
        self.stopped.set()

    def add(self, device, attribute, poll_at, callback=None, returning=True, force_add=False):
        with self.lock:
            if poll_at not in self.polling_schedule:
                self.polling_schedule[poll_at] = OrderedDict()
            if device not in self.polling_schedule[poll_at]:
                self.polling_schedule[poll_at][device] = []
            if device not in self.commands_buffer:
                self.commands_buffer[device] = []

            polling_quadruplet = (attribute, None, callback, returning)
            if force_add or polling_quadruplet not in self.polling_schedule[poll_at][device]:
                self.polling_schedule[poll_at][device].append(polling_quadruplet)

            self.schedule = sorted(self.polling_schedule.keys())

    def remove(self, device, attribute, poll_at, callback=None, returning=True):
        with self.lock:
            try:
                self.polling_schedule[poll_at][device].remove((attribute, None, callback, returning))
                if len(self.polling_schedule[poll_at][device]) == 0:
                    del self.polling_schedule[poll_at][device]
                if len(self.polling_schedule[poll_at]) == 0:
                    del self.polling_schedule[poll_at]

                self.schedule = sorted(self.polling_schedule.keys())
            except (ValueError, KeyError):
                pass

    def stop_polling(self, device):
        with self.lock:
            for polledDevices in self.polling_schedule.values():
                try:
                    if device in polledDevices:
                        del polledDevices[device]
                        self.commands_buffer[device] = []
                except (ValueError, KeyError):
                    pass  # device is no longer within polled devices

    def prepare_command(self, device, command):
        with self.lock:
            self.commands_buffer[device].append(command)

    def prepend_command(self, device, command):
        with self.lock:
            self.commands_buffer[device].insert(0, command)

    def add_connecting_device(self, device):
        with self.lock:
            self.commands_buffer[device] = []
            if device not in self.connecting_devices:
                self.connecting_devices.append(device)

    def remove_connecting_device(self, device):
        with self.lock:
            try:
                self.connecting_devices.remove(device)
            except ValueError:
                pass  # device is no longer within connecting devices

    def check_devices_connection_timeout(self):
        for device in self.connecting_devices[:]:
            if device.connecting_time >= device.connection_timeout:
                self.remove_connecting_device(device)
                device.handle_connection_error(u"Connection timeout")
            elif device.connected or (not device.connecting and not device.closing):
                self.remove_connecting_device(device)
            else:
                device.connecting_time += 1000


class DeviceClientPoller(Thread):
    poll_tick = 2000

    def __init__(self):
        super(DeviceClientPoller, self).__init__()
        self.stopped = Event()
        self.connect_lock = Lock()
        self.commands_buffer = {}
        self.connecting_devices = []
        self.logger = logging.getLogger(u"Console.DeviceClientPoller")

    def run(self):
        self.logger.info(u"Starting devices client poller Thread")
        while not self.stopped.wait(self.poll_tick * 1e-3):
            self.check_devices_connection_timeout()

    def stop(self):
        self.stopped.set()

    def check_devices_connection_timeout(self):
        """
        Check if ZMQ client can communicate with opened port.
        The is that we can connect to the port even if that one is not connected
        on serve side. To check if socket is available we simply try to
        send poll command. If we can get answer proceed with init routine.
        If not, then try it reconnect and try it again.
        :return: None
        """
        for device in self.connecting_devices[:]:
            if self.stopped.is_set():
                break
            if device.poll():
                try:
                    device.handle_connect_event()
                    self.remove_connecting_device(device)
                except DeviceClientError:
                    device.handle_response_error(u"Client initialization failed")
            else:
                if device in self.connecting_devices:
                    device.reconnect()

    def add_connecting_device(self, device):
        with self.connect_lock:
            if device not in self.connecting_devices:
                self.connecting_devices.append(device)

    def remove_connecting_device(self, device):
        with self.connect_lock:
            if device in self.connecting_devices:
                self.connecting_devices.remove(device)


class DeviceAsyncoreLoop(Thread):
    dummy_device = None

    def __init__(self):
        super(DeviceAsyncoreLoop, self).__init__()
        self.dummy_device = asynchat.async_chat()
        self.dummy_device.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.dummy_device.socket.setblocking(False)
        self.setDaemon(True)

    def run(self):
        asyncore.loop(use_poll=True, timeout=0.1)

    def stop(self):
        if self.is_alive():
            self.dummy_device.close()

        asyncore.close_all()


control_asyncore_loop = DeviceAsyncoreLoop()
control_device_poller = DevicePoller()
control_port_read_loop = PortReadLoop()
control_device_client_poller = DeviceClientPoller()



import json
import time
from threading import Thread, Event, Lock

import zmq

import kamzik3
from kamzik3 import SEP, DeviceClientError
from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetDataAccess import get_from_dict, set_in_dict
from kamzik3.snippets.snippetsControlLoops import control_device_client_poller
from kamzik3.snippets.snippetsJson import JsonKamzikEncoder

""" Example of yaml configuration
DeviceClient0: &DeviceClient0 !Device:kamzik3.devices.deviceClient.DeviceClient
    host: &Server 127.0.0.1
    port: &Port 60000
    device_id: DeviceIdOnServer
"""


class DeviceClient(Device):
    """
    : connect()
    : handle_connect_event()
            : handle_connect()
            : handle_configuration_event()
                : handle_configuration()
    """
    initial_status = STATUS_DISCONNECTED

    def __init__(self, host, port, device_id=None, config=None):
        Device.__init__(self, device_id, config)
        self.client_comm_lock = Lock()
        self.device_poller = control_device_client_poller
        self.host = host
        self.port = port
        self.removing = False
        self.subscriber_port = None
        self.socket = None
        self.subscriber_socket = None
        self.callbacks = {}
        self.stopped = Event()
        self.subscribe_starts_with_list = []
        self.connect()

    def connect(self):
        """
        Here we want to connect first to ZMQ main server.
        We have to verify connection and do it in non blocking fashion.
        We just push devices in a pool and verify it in device_poller Thread.
        :return:
        """
        self.logger.info(u"Initiating connection to the server {}:{}".format(self.host, self.port))
        self.connecting = True
        self.connected = False
        self.set_status(STATUS_CONNECTING)
        # Get REQ type of socket
        self.socket = zmq.Context.instance().socket(zmq.REQ)
        # Set receive timeout to 3000 ms
        self.socket.setsockopt(zmq.RCVTIMEO, 3000)
        # Don't let socket linger after close
        # This is VERY IMPORTANT not to set to anything but 0
        self.socket.setsockopt(zmq.LINGER, 0)
        try:
            self.socket.connect("tcp://{}:{}".format(self.host, self.port))
            # Add socket to connection loop and return from connect method
            self.device_poller.add_connecting_device(self)
        except zmq.ZMQError:
            self.handle_connection_error(u"Fatal connection error")
            return

    def handle_connect_event(self):
        """
        This method is called after we verified connection with main server.
        We continue with init method to subscribe to devices publisher.
        :return: None
        """
        if self.init():
            self.connected = True
            self.connecting = False
            self.handle_configuration_event()

            self.set_status(self.get_attribute([ATTR_STATUS, VALUE]))

            # Subscribe for all attribute changes
            token = self.get_token(TOKEN_ATTRIBUTE)
            self._subscribe(token)
            self.subscribe_starts_with_list = token

            # Check for attribute sharing and subscribe to them
            if len(self.attributes_sharing_map) > 0:
                self.subscribe_starts_with_list = [self.subscribe_starts_with_list]
                for device_id, groups in self.attributes_sharing_map.items():
                    token = self.get_token(TOKEN_ATTRIBUTE, device_id)
                    for source_group, (target_group, source_attributes) in groups.items():
                        if source_group != "null":
                            sub_token = ".".join((token, source_group))
                            self.subscribe_starts_with_list.append("{}.{}".format(token, source_group))
                        else:
                            sub_token = token
                            self.subscribe_starts_with_list.append(sub_token)
                        if source_attributes == ALL:
                            self._subscribe(sub_token)
                        else:
                            for source_attribute in source_attributes:
                                self._subscribe(".".join((sub_token, source_attribute)))
                        self.subscribe_starts_with_list.append(sub_token)
                self.subscribe_starts_with_list = tuple(self.subscribe_starts_with_list)

    def reconnect_allowed(self):
        if self.connected:
            return False
        else:
            return self.allow_reconnect

    def _init_subscriber_thread(self):
        """
        We are connected to the main server.
        Now we want to connect to the devices publisher.
        :return: None
        """
        self.logger.info(u"Initializing {} client subscriber thread".format(self.device_id))
        # Get SUB type of socker
        self.subscriber_socket = zmq.Context.instance().socket(zmq.SUB)
        # Set receive timeout to 1000 ms
        self.subscriber_socket.setsockopt(zmq.RCVTIMEO, 2000)
        # Don't let socket linger after close
        # This is VERY IMPORTANT not to set to anything but 0
        self.subscriber_socket.setsockopt(zmq.LINGER, 0)
        self.subscriber_socket.connect("tcp://{}:{}".format(self.host, self.subscriber_port))
        # Start separate Thread to read data from publisher
        Thread(target=self._subscriber_thread).start()

    def _subscriber_thread(self):
        """
        Continuously collect data from publisher.
        Interrupt every zmq.RCVTIMEO and check if client is connected.
        :return: None
        """
        while not self.stopped.isSet():
            try:
                token, data = self.subscriber_socket.recv().decode().split(SEP)
                data = json.loads(data)
                self.handle_readout(token, data)
            except zmq.Again:
                # This exception is raised due mq.RCVTIMEO set to 1000ms.
                # We interrupt recv loop to check if devices is still connected
                # If not socket will be cleanly closed
                if self.connected and not self.client_comm_lock.locked():
                    self.poll()
            except AttributeError:
                pass  # Socket is not defined

        self.handle_closing_subscriber()

    def _subscribe(self, topic):
        """
        Subscribe for Topic.
        :param topic: string
        :return: None
        """
        self.subscriber_socket.setsockopt_string(zmq.SUBSCRIBE, topic)

    def _unsubscribe(self, topic):
        """
        Unsubscribe Topic.
        :param topic: string
        :return: None
        """
        self.subscriber_socket.setsockopt_string(zmq.UNSUBSCRIBE, topic)

    def set_callback(self, token, callback):
        """
        Set callback when token is received from publisher.
        If remove_after is set, remove callback from buffer and don't repeat it again.
        Callback has to be callable function.
        :param token: string
        :param callback: function(devices, data)
        :return: bool
        """
        assert callable(callback)

        if token not in self.callbacks:
            self.callbacks[token] = []

        callback_data = callback

        if callback_data not in self.callbacks[token]:
            self.callbacks[token].append(callback_data)
            self._subscribe(token)
            self.logger.debug(u"Set callback for token {}".format(token))
            return token
        else:
            self.logger.warning(u"Callback is already set")
            return False

    def handle_readout_callback(self, token, callback_data):
        """
        Handle callback for token.
        :param token: string
        :param callback_data: mixed
        :return: None
        """
        self._unsubscribe(token)
        while len(self.callbacks.get(token)) > 0:
            callback = self.callbacks.get(token).pop()
            callback(*callback_data)

        if len(self.callbacks[token]) == 0:
            del self.callbacks[token]

    def handle_readout(self, token, data):
        if token.startswith(self.subscribe_starts_with_list):
            source_device_id = token.split(".", 1)[0]
            if source_device_id != self.device_id:
                group = data[0][:-1][0]
                try:
                    (target_group, shared_attributes) = self.attributes_sharing_map[source_device_id][group]
                except KeyError:
                    group = "null"
                    (target_group, shared_attributes) = self.attributes_sharing_map[source_device_id][group]
                if group is "null" and target_group is not "null":
                    data[0].insert(0, target_group)
                elif group is not "null" and target_group is "null":
                    del data[0][0]
                else:
                    data[0][0] = target_group
            self._set(*data)
        elif token in self.callbacks:
            self.handle_readout_callback(token, data)

    def init(self):
        """
        Initiate devices connection.
        Get publisher port and devices attributes.
        :return:
        """
        try:
            with self.client_comm_lock:
                self.socket.send(SEP.join((self.device_id, INIT)).encode())
                status, token, response = self.handle_server_response()

            if status == RESPONSE_OK:
                self.subscriber_port, attributes, self.attributes_sharing_map, self.exposed_methods, self.qualified_name = json.loads(
                    response)
                self.initial_status = attributes[ATTR_STATUS][VALUE]
                self._sync_attributes(attributes)
                for method_name, attributes in self.exposed_methods[:]:
                    setattr(self, method_name,
                            lambda method_name=method_name, **kwargs: self.method(method_name, kwargs))
                self._init_subscriber_thread()
                return True
            else:
                raise DeviceClientError(u"Failed to initialize device")

        except (zmq.Again, zmq.ZMQError):
            self.handle_response_error(u"Init error")
            self.reconnect()
            return False

    def _sync_attributes(self, synced_attributes):
        """
        synchronize new attributes with already existing attributes.
        This is important to prevent creating new attribute objects.
        We can just create new attributes, but we would lost all callbacks and all references will become invalid.
        :param synced_attributes: dict
        :return:
        """
        # Walk thru all items of dictionary
        for key, attribute in synced_attributes.items():
            self._sync_attribute(key, attribute)

    def _sync_attribute(self, key, attribute):
        """
        synchronize one attribute from new dict to existing self.attributes
        :param key: string
        :param attribute: dict
        :return:
        """
        if key in self.attributes:
            if Attribute.is_attribute(attribute):
                if key == ATTR_STATUS:
                    return
                # Key exists in self.attributes
                # Prevent attribute changes when syncing with new values
                # with self.attributes[key].update_lock:
                for attr_key in attribute.keys():
                    if self.attributes[key][attr_key] != attribute[attr_key]:
                        self.attributes[key][attr_key] = attribute[attr_key]
            else:
                # Attribute is only sub group, continue syncing deeper
                self._sync_attributes(attribute)
        else:
            # Key des not exists in self.attributes
            if Attribute.is_attribute(attribute):
                # Create completely new Attribute under specified Key
                self.attributes[key] = Attribute.from_dict(attribute)
            else:
                # Attribute is only sub group, continue syncing deeper
                self.attributes[key] = Attribute.from_dict(attribute)

    def get_token(self, topic, device_id=None):
        return "{}.{}".format(self.device_id if device_id is None else device_id, topic)

    def set_attribute(self, attribute, value, callback=None):
        """
        Set attribute on devices server.
        :param attribute:
        :param value:
        :param callback:
        :return:
        """
        if not self.connected or self.closing:
            raise DeviceClientError(
                u"Remote set attribute {} to value {} failed. Device client is not connected".format(attribute, value))
        try:
            attribute_value = json.dumps((attribute, value), ensure_ascii=True)
            with self.client_comm_lock:
                self.socket.send(SEP.join((self.device_id, SET, attribute_value)).encode())
                status, token, response = self.handle_server_response()
            if status == RESPONSE_OK:
                if callback is not None:
                    if token:
                        self.set_callback(self.get_token(token), callback)
                    else:
                        callback(attribute, value)
            else:
                raise DeviceClientError(
                    u"Failed to set remote attribute {} to value {}\n{}".format(attribute, value, response))

            return response

        except (zmq.Again, zmq.ZMQError):
            self.handle_response_error(u"Set error")
            return False

    def _set(self, attribute, value):
        """
        This sets Device attribute.
        Use this function when You want to set attribute by tuple or list key.
        Example: Device.set((ATTR_STATUS, VALUE), STATUS_IDLE)
        Attribute value is pushed into server if Device is connected on any.
        To reduce amount of pushed attributes we check if value is different from previous one.
        :param attribute: tuple, list, str
        :param value: mixed
        :return: None
        """
        current_value = get_from_dict(self.attributes, attribute)
        if value == current_value:
            return
        set_in_dict(self.attributes, attribute, value)

    def poll(self):
        """
        Poll devices for activity.
        :return:
        """
        if not self.connected and not self.connecting or not self.socket:
            raise DeviceClientError(
                u"Remote server poll failed. Device client is not connected")
        try:
            with self.client_comm_lock:
                self.socket.send(SEP.join((self.device_id, POLL)).encode())
                status, token, response = self.handle_server_response()
            if status == RESPONSE_OK:
                return response
            else:
                raise DeviceClientError(u"Failed to poll remote device")

        except zmq.Again:
            return False
        except zmq.ZMQError as e:
            self.handle_response_error(u"Poll error: {}".format(e))
            return False

    def command(self, command, callback=None):
        """
        Execute raw command directly on devices.
        :param command: string
        :param callback: function(devices, outputOfCommand)
        :return:
        """
        if not self.connected or self.closing:
            raise DeviceClientError(
                u"Remote command {0!r} execution failed. Client is not connected".format(command))

        try:
            with self.client_comm_lock:
                self.socket.send(SEP.join((self.device_id, COMMAND, command, "1" if callback else "0")).encode())
                status, token, response = self.handle_server_response()

            if status == RESPONSE_OK:
                if token and callback is not None:
                    self.set_callback(self.get_token(token), callback)
            else:
                raise DeviceClientError(u"Failed to execute remote command '{}'\n{}".format(command, response))

            return response
        except (zmq.Again, zmq.ZMQError) as e:
            self.handle_response_error(u"Command error: {}".format(e))
            return False

    def method(self, method, attributes=None):
        """
        Execute devices method with associated attributes.
        :param method: string
        :param attributes: json formatted string
        :return:
        """
        if not self.connected or self.closing:
            raise DeviceClientError(
                u"Method {} remote execution failed. Client is not connected".format(method))

        try:
            dumped_attributes = json.dumps(attributes, ensure_ascii=True, cls=JsonKamzikEncoder)
            with self.client_comm_lock:
                self.socket.send(SEP.join((self.device_id, METHOD, str(method), dumped_attributes)).encode())
                status, token, response = self.handle_server_response()
            if status == RESPONSE_OK:
                return json.loads(response)
            else:
                raise DeviceClientError(u"Failed to execute remote method '{}'\n{}".format(method, response))

        except (zmq.Again, zmq.ZMQError) as e:
            self.handle_response_error(u"Method error: {}".format(e))
            return False

    def handle_server_response(self):
        '''
        This is crucial moment here.
        If server takes more then RCVTIMEO time, then we get zmq error.
        In that case modify server response time or increase RCVTIMEO time.
        '''
        response = self.socket.recv().decode()
        status, token, response = response.split(SEP)
        return status, token, response

    def handle_configuration(self):
        start_at = time.time()
        self._config_attributes()
        self._config_commands()
        self.set_status(self.initial_status)
        self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

    def handle_closing_subscriber(self):
        if self.subscriber_socket:
            self.subscriber_socket.close()
        if self.logger is not None:
            self.logger.info(u"Client closed")
        if not self.removing:
            self.set_status(STATUS_DISCONNECTED)
            self.closing = False
            self.connected = False

            if self.response_error or self.connection_error:
                self.reconnect()

    def handle_response_error(self, message=None):
        self.logger.error(message)
        self.response_error = True
        self.close()

    def handle_connection_error(self, message=None):
        self.logger.error(message)
        self.connection_error = True
        self.disconnect()

    def handle_command_error(self, readout_command, readout_output):
        self.logger.error(
            u"Command error\nCommand: {}\nOutput: {}".format(readout_command, readout_output))

    def disconnect(self):
        if self.connecting:
            # Device is connecting, but here we want to disconnect it.
            # To prevent automatic reconnection, just remove it from connecting devices.
            self.device_poller.remove_connecting_device(self)
        else:
            # Device is fully connected and it's not in connecting loop
            # Therefor we can safely close it
            self.close()

    def close(self):
        self.logger.info(u"Closing client")
        self.closing = True
        self.connected = False
        self.set_status(STATUS_DISCONNECTING)
        if self.socket:
            self.socket.close()
        self.stopped.set()
        # Here we doing nothing more
        # After stopped flag is set handle_closing_subscriber is fired and finish
        # client closing procedure

    def remove(self):
        self.removing = True
        Device.remove(self)



from PyQt5.QtCore import pyqtSlot, pyqtSignal
from PyQt5.QtWidgets import QWidget, QLabel, QComboBox
from pint import UndefinedUnitError

from kamzik3 import units
from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.snippets.snippetsUnits import get_attribute_unit_range


class AttributeWidget(QWidget):
    input_widget = None
    unit_widget = None
    unit = None
    font_color = u"black"
    background_color = u"white"
    attribute = None
    attribute_type_cast = str
    sig_unit_changed = pyqtSignal("QString", "QString")

    def __init__(self, attribute, config=None, parent=None):
        assert isinstance(attribute, Attribute)
        self.attribute = attribute
        self.config = config
        if self.config is None:
            self.config = {}
        QWidget.__init__(self, parent=parent)
        self.setupUi()

    def _get_static_unit_widget(self):
        display_uint_name = self.attribute[UNIT]
        if display_uint_name == u"percent":
            display_uint_name = u"%"
        return QLabel(u" {}".format(display_uint_name))

    def _get_combox_unit_widget(self):
        unit_combobox = QComboBox()
        unit_combobox.addItems(
            get_attribute_unit_range(self.attribute)
        )

        def index_changed(_):
            self.set_unit(str(unit_combobox.currentText()))

        unit_combobox.currentIndexChanged.connect(index_changed)
        unit_combobox.wheelEvent = lambda *args: None
        index = unit_combobox.findText("{:~}".format(units.Unit(self.attribute[UNIT])))
        if index == -1:
            unit_combobox.setCurrentIndex(0)
        else:
            unit_combobox.setCurrentIndex(index)
        return unit_combobox

    def _set_unit_widget(self):
        self.unit = self.attribute[UNIT]
        if self.unit is not None:
            if self.unit in units.derived_units or self.config.get("static_unit", False):
                self.unit_widget = self._get_static_unit_widget()
            else:
                try:
                    self.unit_widget = self._get_combox_unit_widget()
                except (UndefinedUnitError, AttributeError, KeyError):
                    self.unit_widget = self._get_static_unit_widget()
        else:
            self.unit_widget = QLabel(u"")

    def set_tooltip(self, value=None):
        tooltip_value = self.attribute[DESCRIPTION]
        if tooltip_value is not None:
            tooltip_value += "\n" + value
        else:
            tooltip_value = value

        if tooltip_value is not None:
            self.input_widget.setToolTip(tooltip_value)

    def update_value(self):
        self.set_value(self.attribute.value())

    @pyqtSlot("PyQt_PyObject")
    def _set_input_widget(self):
        raise NotImplementedError("Has to be implemented")

    def get_attribute_value(self):
        raise NotImplementedError("Has to be implemented")

    def get_widget_value(self):
        raise NotImplementedError("Has to be implemented")

    def set_value(self, value):
        raise NotImplementedError("Has to be implemented")

    @pyqtSlot("QString")
    def set_unit(self, unit=None, reset=False):
        raise NotImplementedError("Has to be implemented")

    def get_unit(self):
        return self.unit

    def setDisabled(self, bool):
        self.input_widget.setDisabled(True)
        super().setDisabled(bool)

    def setupUi(self, parent=None):
        self.unit = self.attribute[UNIT]
        self._set_input_widget()
        self._set_unit_widget()

        if self.attribute[VALUE] is not None:
            self.set_value(self.attribute.value())

        if self.config is not None and self.config.get("unit") is not None:
            self.unit_widget.setCurrentText(self.config.get("unit"))
        self.input_widget.__dict__["update_value"] = self.update_value

    def close(self):
        if self.input_widget is not None:
            self.input_widget.setParent(None)
            del self.input_widget
        if self.unit_widget is not None:
            self.unit_widget.setParent(None)
            del self.unit_widget

        self.attribute = None
        self.config = None
        return super().close()



from PyQt5.QtCore import pyqtSlot
from PyQt5.QtWidgets import QSizePolicy

from kamzik3.constants import *
from kamzik3.gui.attributeWidgets.attributeWidget import AttributeWidget
from kamzik3.snippets.snippetsWidgets import CustomLineEdit


class AttributeStringWidget(AttributeWidget):

    def _set_input_widget(self):
        self.input_widget = CustomLineEdit()
        if self.attribute[READONLY]:
            self.input_widget.setReadOnly(True)
        else:
            self.input_widget.setStyleSheet(".CustomLineEdit {background:#e6ffd9}")
            # self.input_widget.returnPressed.connect(
            #     lambda attribute=self.input_widget: self.set_value(attribute.text()))
            if self.attribute[DESCRIPTION] is not None:
                self.input_widget.setToolTip(self.attribute[DESCRIPTION])

        self.input_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

    def set_value(self, value):
        self.input_widget.setText(str(value))

    def get_attribute_value(self):
        return self.input_widget.text()

    def get_widget_value(self):
        return self.input_widget.text()

    @pyqtSlot("QString")
    def set_unit(self, unit=None, reset=False):
        return



import logging
import sys

# Base formatter for logging.
import traceback
from logging.handlers import TimedRotatingFileHandler

base_log_formatter = logging.Formatter('%(asctime)s, %(name)s, %(levelname)s, %(message)s', datefmt='%Y-%m-%d %H:%M:%S')


def set_file_handler(logger, log_output_dir, formatter=base_log_formatter):
    handler = logging.FileHandler(log_output_dir)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return handler


def set_rotating_file_handler(logger, log_output_dir, formatter=base_log_formatter):
    handler = TimedRotatingFileHandler(log_output_dir, when="midnight")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return handler


def get_console_handler(log_level=logging.DEBUG, formatter=base_log_formatter):
    console_handler = logging.StreamHandler()
    console_handler.setLevel(log_level)
    console_handler.setFormatter(formatter)
    return console_handler


def set_sys_exception_handler(logger):
    def handle_exception(exc_type, exc_value, exc_traceback):
        logger.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))
        sys.__excepthook__(exc_type, exc_value, exc_traceback)

    sys.excepthook = handle_exception


def print_exception():
    print(traceback.format_exc())



import json
import logging
from collections import deque
from threading import Thread, Event

import zmq

import kamzik3
from kamzik3 import SEP, CommandFormatException, DeviceServerError, DeviceError
from kamzik3.constants import *
from kamzik3.snippets.snippetDataAccess import rgetattr
from kamzik3.snippets.snippetsJson import JsonPintHook
from kamzik3.snippets.snippetsYaml import YamlSerializable


class DeviceServer(YamlSerializable):
    publisher_tick = 10  # ms
    server_thread = None

    def __init__(self, host, server_port, publisher_port):
        self.logger = logging.getLogger(u"Console.DeviceServer@{}:{}".format(host, server_port))
        self.host = host
        self.server_port = server_port
        self.publisher_port = publisher_port
        self.messages_queue = deque()
        self.devices = {}
        self.close_event = Event()
        self._init_publisher_thread()
        self._init_server_thread()
        kamzik3.session.set_device_server(self)

    def _init_server_thread(self):
        """
        Initialize connection to Device server as a client.
        Connection is done via ZMQ REP socket.
        :return: None
        """
        self.logger.info(u"Initializing server thread")
        self.server_socket = zmq.Context.instance().socket(zmq.REP)
        # Set receive timeout to 1000 ms
        self.server_socket.setsockopt(zmq.RCVTIMEO, 1000)
        # Don't let socket linger after close
        # This is VERY IMPORTANT not to set to anything but 0
        self.server_socket.setsockopt(zmq.LINGER, 0)
        self.server_socket.bind("tcp://{}:{}".format(self.host, self.server_port))
        self.server_thread = Thread(target=self._server_thread)
        self.server_thread.start()

    def _server_thread(self):
        while not self.close_event.isSet():
            try:
                data = self.server_socket.recv().decode().split(SEP)
                device_id, instruction = data[0], data[1]

                # Execute command directly on devices
                if instruction == COMMAND:
                    command, with_token = data[2], int(data[3])
                    token = self.get_device(device_id).command(command, with_token=int(with_token))
                    status, response = RESPONSE_OK, len(data[3])

                # Get devices attribute
                elif instruction == GET:
                    attribute = json.loads(data[2], object_hook=JsonPintHook)
                    response = self.get_device(device_id)._get(attribute)
                    response = json.dumps(response, ensure_ascii=True)
                    status, token = RESPONSE_OK, 0

                # Set devices attribute
                elif instruction == SET:
                    attribute, attribute_value = json.loads(data[2], object_hook=JsonPintHook)
                    set_token = self.get_device(device_id).set_attribute(attribute, attribute_value)
                    status, token, response = RESPONSE_OK, set_token, len(data[2])

                # Execute devices method with attributes
                elif instruction == METHOD:
                    method, attributes = data[2], json.loads(data[3], object_hook=JsonPintHook)
                    device = self.get_device(device_id)
                    response = rgetattr(device, method)(**attributes)
                    response = json.dumps(response, ensure_ascii=True)
                    status, token = RESPONSE_OK, 0

                # Poll devices for activity
                elif instruction == POLL:
                    status, token, response = RESPONSE_OK, 0, 1

                # Init devices
                elif instruction == INIT:
                    device = self.get_device(device_id)

                    if device is None:
                        raise DeviceServerError(
                            "Device {} is not registered on the server or publisher is not ready".format(device_id))
                    elif self.publisher_port is None:
                        raise DeviceServerError("Server publisher is not ready")
                    elif not device.in_statuses(READY_DEVICE_STATUSES):
                        raise DeviceServerError("Device is not ready")
                    else:
                        response = json.dumps(
                            (self.publisher_port, device.attributes, device.attributes_sharing_map,
                             device.exposed_methods, device.qualified_name),
                            ensure_ascii=True)
                        status, token = RESPONSE_OK, 0

                # None of above, request not implemented
                else:
                    status, token, response = RESPONSE_ERROR, 0, len(data[3])

                self.server_socket.send(SEP.join((str(status), str(token), str(response))).encode())
            except zmq.Again:
                """
                This exception is raised due mq.RCVTIMEO set to 1000ms.
                We interrupt recv loop to check if devices is still connected
                If not socket will be cleanly closed
                """
                pass
            except CommandFormatException:
                status, token, response = RESPONSE_ERROR, 0, u"Command format error"
                self.server_socket.send(SEP.join((str(status), str(token), str(response))).encode())
            except DeviceServerError:
                status, token, response = RESPONSE_ERROR, 0, u"Device server error"
                self.server_socket.send(SEP.join((str(status), str(token), str(response))).encode())
            except DeviceError as e:
                status, token, response = RESPONSE_ERROR, 0, u"Command error: {}".format(e)
                self.server_socket.send(SEP.join((str(status), str(token), str(response))).encode())
        self.handle_closing_server()

    def _init_publisher_thread(self):
        self.logger.info(u"Initializing publisher thread")
        self.publisherSocket = zmq.Context.instance().socket(zmq.PUB)
        self.publisherSocket.bind("tcp://{}:{}".format(self.host, self.publisher_port))
        Thread(target=self._publisher_thread).start()

    def _publisher_thread(self):
        while not self.close_event.wait(self.publisher_tick * 1e-3):
            while self.messages_queue:
                header, message = self.messages_queue.popleft()
                try:
                    out_message = header + SEP + json.dumps(message, ensure_ascii=True)
                    self.publisherSocket.send(out_message.encode())
                except TypeError:
                    print(message)

    def add_device(self, device):
        self.logger.info(u"Adding {} into server".format(device.device_id))
        if device.device_id not in self.devices:
            self.devices[device.device_id] = device
            device.set_device_server(self)
            return True
        else:
            return False

    def remove_device(self, device):
        self.logger.info(u"Removing {} from server".format(device.device_id))
        try:
            del self.devices[device.device_id]
            device.clear_device_server()
            return True
        except ValueError:
            return False

    def get_device(self, device_id):
        return self.devices.get(device_id, None)

    def push_message(self, device_id, message, token=None):
        try:
            if token:
                header = "{}.{}".format(device_id, token)
            else:
                header = device_id
            self.messages_queue.append((header, message))
        except TypeError:
            self.logger.exception(u"Error pushing message to client")

    def handle_closing_server(self):
        if self.server_socket:
            self.server_socket.close()

    def close(self):
        self.logger.info(u"Closing server")
        self.close_event.set()
        if self.server_thread is not None:
            self.server_thread.join(1)
